#    Copyright 2010 Benet Devereux
#    This file is part of batholith.
#    The program 'batholith' is free software: you can redistribute it and/or modify
#     it under the terms of the GNU Lesser General Public License as published by
#     the Free Software Foundation, either version 3 of the License, or
#     (at your option) any later version.
# 
#     This program is distributed in the hope that it will be useful,
#     but WITHOUT ANY WARRANTY; without even the implied warranty of
#     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#     GNU General Public License for more details.
# 
#     You should have received a copy of the GNU Lesser General Public License
#     along with this program.  If not, see <http://www.gnu.org/licenses/>.

import logging, math, os, pygame, random, sys
from geom import *
from pygame.locals import * 

debugMode = False

# Constants
#
## App state constants
STARTSCREEN = 0
MAINSCREEN = 1
ENDGAMESCREEN = 2
FAREWELLSCREEN = 3
## Event constants
FAILEDMOVEEVENT = pygame.USEREVENT + 1
CAVEGROWTHEVENT = FAILEDMOVEEVENT + 1
REPEATEDMOVEEVENT = CAVEGROWTHEVENT + 1
CAVEINEVENT = REPEATEDMOVEEVENT + 1
DEATHEVENT = CAVEINEVENT + 1
HUNGEREVENT = DEATHEVENT + 1    # even without moving we get hungry...

def initialize_fonts():
    pygame.font.init()
    global titleFont, infoFont, regularFont, detailsFont
    titleFont = pygame.font.SysFont('couriernew', 80)
    infoFont = pygame.font.SysFont('couriernew', 20)
    regularFont = pygame.font.SysFont('couriernew', 10)
    detailsFont = pygame.font.SysFont('couriernew', 13)
    
def initialize_images():
    global copyleftImage
    copyleftImage = pygame.image.load(os.path.join('data' ,'cc-LGPL-a.png'))

def initialize_sound():
    pygame.mixer.init()
    global bonkSound, caveinSound, chimeSound
    bonkSound = pygame.mixer.Sound('data/rock2.wav')
    caveinSound = pygame.mixer.Sound('data/cavein.wav')
    chimeSound = pygame.mixer.Sound('data/chime.wav')
    
def initialize_colours():
    global white, c_rock, c_stable, c_apical_visible, c_apical_invisible, c_unstable_visible, c_unstable_invisible
    white = pygame.color.Color('white')
    c_rock = pygame.color.Color('black')
    c_stable = pygame.color.Color('red')
    c_apical_visible = pygame.color.Color('green')
    if debugMode:
        c_apical_invisible = c_apical_visible 
    else:
        c_apical_invisible  = c_rock
    if debugMode:
        c_unstable_invisible = pygame.color.Color(10,10,10,0)
    else:
        c_unstable_invisible = c_rock
    c_unstable_visible = pygame.color.Color(20,20,20,0)

def initialize_text():
    global title, gameOverText, creditText, pressToPlayText, instructionsTextList, soundCreditsTextList
    title = titleFont.render('batholith', False, white)
    gameOverText = titleFont.render('game over', False, white)
    creditText = detailsFont.render('Written for PyWeek #10 by Benet Devereux (benetdevereux@gmail.com)', False, white)
    pressToPlayText = infoFont.render('... press any key to play ...', False, c_stable)
    instructionsTextList = [detailsFont.render(t, False, white) for t in ['Move with cursor keys', 'Use . to pick things up',
         'Be careful: this can cause a cave-in!', 'Collect gems (*)', 'Use mushrooms (m) to restore health', 'Space key pages through status updates', 'q to quit']]
    soundCreditsTextList = [detailsFont.render(t, False, white) for t in ['Sound Samples Used',
    'By acclivity (http://www.freesound.org/usersViewSingle.php?id=37876)',
    'Goblet_G_Loud.wav (http://www.freesound.org/samplesViewSingle.php?id=30602)',
    'By Benboncan (http://www.freesound.org/usersViewSingle.php?id=634166)',
    'Rockfall in mine.wav (http://www.freesound.org/samplesViewSingle.php?id=60085)', 
    'By klankschap (http://www.freesound.org/usersViewSingle.php?id=34216)', 
    'bonk.wav (http://www.freesound.org/samplesViewSingle.php?id=28077)']]

def initialize_loggers():
    global caveEventLogger, playerEventLogger
    # Set up some loggers
    caveEventLogger = logging.getLogger("cave events")     # events to do with the cave system
    playerEventLogger = logging.getLogger("player events") # events to do with player actions
    ceLogger = logging.StreamHandler()
    peLogger = logging.StreamHandler()
    caveEventLogger.addHandler(ceLogger)
    playerEventLogger.addHandler(peLogger)
    if debugMode:
        lvl = logging.DEBUG
    else:
        lvl  = logging.INFO
    caveEventLogger.setLevel(lvl)
    playerEventLogger.setLevel(lvl)

def initialize(x,y):
    global gameState, window, screen, objectsurfaces, winWidth, winHeight, statusLineHeight
    initialize_fonts()
    initialize_images()
    initialize_sound()
    initialize_colours()
    initialize_text()
    initialize_loggers()
    gameState = STARTSCREEN
	# Surfaces for objects and other constants
    objectsurfaces = { '@' : regularFont.render('@', False, white), '*' : regularFont.render('*', False, white), 'm': regularFont.render('m', False, white) }
    statusLineHeight = infoFont.size("Py")[1] + 10
    (winWidth, winHeight) = (x/objectsurfaces['@'].get_width(), (y - statusLineHeight)/objectsurfaces['@'].get_height())
    window = pygame.display.set_mode((x,y))
    pygame.display.set_caption("Batholith: PyWeek 10 Challenge Entry")
    screen = pygame.display.get_surface()

initialize(800,600)

# Determine size of grid and height of status line


class GridManager(object):
    def __init__(self, w, h):
        self.w = w
        self.h = h
        self.grid = []
        for i in range(h):
            self.grid.append([None] * w)
    def run_initializer(self, f):
        for y in range(self.h):
            for x in range(self.w):
                self.grid[y][x] = f(x,y)
    def in_bounds(self,x,y):
        return x >= 0 and x < self.w and y >= 0 and y < self.h
    def __call__(self, *listargs):
        return self.grid[listargs[1]][listargs[0]]
    def __repr__(self):
        return "%d x %d gridmanager" % (self.w, self.h)

class GridElementState(object):
    ROCK = 0
    STABLE = 1
    UNSTABLE = 2
    APICAL = 3
    def __init__(self, state, x, y, direction_vector = None, visible = False):
        self.state = state
        self.x = x
        self.y = y
        self.direction_vector = direction_vector
        self.visible = visible
        self.contents = []
    def try_get(self):
        retVal = None
        try:    
            retVal = self.contents.pop()
        except IndexError:
            retVal = None
        return retVal
    def __repr__(self):
        return "%s at (%d, %d)" % (['rock', 'stable', 'unstable', 'apical'][self.state], self.x, self.y)

class CaveGrowthManager(object):
    def __init__(self, gridmanager, pr_treasure=0.05, pr_apicalGrow= 0.5, pr_apicalSplit= 0.05, pr_caveIn=0.5):
        self.gridmanager = gridmanager
        self.pr_treasure = 0.05
        self.pr_apicalGrow = pr_apicalGrow
        self.pr_apicalSplit = pr_apicalSplit
        self.pr_caveIn = pr_caveIn              # chance that player action causes a cave-in
        self.pr_mushroomGrow = 0.05
        self.apicals = []
        self.unstables = []
    def starting_point(self, size):
        (midX, midY) = (self.gridmanager.w / 2, self.gridmanager.h / 2)
        for x in range(midX - size/2, midX + size/2):
            for y in range(midY - size/2 , midY + size/2):
                self.gridmanager(x,y).state = GridElementState.STABLE
        firstApical = self.gridmanager(random.randrange(midX-size/2, midX+size/2), midY + size/2)
        firstApical.state = GridElementState.APICAL
        firstApical.direction_vector = [0,1]
        self.apicals.append(firstApical)
        secondApical = self.gridmanager(random.randrange(midX-size/2, midX+size/2), midY - size/2 - 1)
        secondApical.state = GridElementState.APICAL
        secondApical.direction_vector = [0,-1]
        self.apicals.append(secondApical)
    def expand_crack(self, x0, y0, x1, y1, dirvect):
        self.gridmanager(x1,y1).state = GridElementState.APICAL
        self.gridmanager(x1,y1).direction_vector = dirvect
        self.gridmanager(x0,y0).state = GridElementState.UNSTABLE        
        if random.random() < self.pr_treasure:
            self.gridmanager(x1,y1).contents.append('*')
            caveEventLogger.debug("treasure drop @ %d, %d" % (x1,y1))
        if random.random() < self.pr_mushroomGrow:
            self.gridmanager(x1, y1).contents.append('m')
    def iterate_over_unstables(self):
        if self.unstables and random.random() < self.pr_caveIn:
            victim = self.unstables[random.randrange(len(self.unstables))]
            victim.state = GridElementState.ROCK
            pygame.event.post(pygame.event.Event(CAVEINEVENT, { 'x': victim.x, 'y': victim.y }))
    def iterate_over_apicals(self):
        if self.apicals == []:     # handle the case where all the apicals die; pick a new 'random' unstable
            newapical = self.unstables[random.randrange(len(self.unstables))]
            self.unstables.remove(newapical)
            self.apicals.append(newapical)
        for a in self.apicals:
            caveEventLogger.debug( "Will apical at %d, %d grow?" % (a.x, a.y))
            r = random.random()
            if r <= self.pr_apicalGrow:
                caveEventLogger.debug( "Yes!")
                nextX = a.x + a.direction_vector[0]
                nextY = a.y + a.direction_vector[1]
                current = None
                # Case 0: going off edge
                if not self.gridmanager.in_bounds(nextX, nextY):
                    caveEventLogger.debug( "Going off edge")
                    a.state = GridElementState.UNSTABLE
                    self.apicals.remove(a)
                    return
                else:
                    next = self.gridmanager(nextX, nextY)
                # Case 1: next location is rock.
                if next.state == GridElementState.ROCK:
                    caveEventLogger.debug( "Moving into rock!")
                    self.expand_crack(a.x, a.y, nextX, nextY, a.direction_vector)
                    self.apicals.remove(a)
                    self.unstables.append(a)
                    self.apicals.append(next)
                # Case 2: next location is stable. Current becomes unstable.
                elif next.state == GridElementState.STABLE:
                    caveEventLogger.debug( "Opening into stable area.")
                    a.state = GridElementState.UNSTABLE
                    self.apicals.remove(a)
                # Case 3: next is unstable; again, just open into it.
                elif next.state == GridElementState.UNSTABLE:
                    caveEventLogger.debug( "Opening into unstable area.")
                    a.state = GridElementState.UNSTABLE
                    self.apicals.remove(a)
            else:
                if r <= self.pr_apicalGrow + self.pr_apicalSplit:
                    caveEventLogger.debug( "It will split")
                    leftBranch = rotate_left(a.direction_vector)
                    (nextXLeft, nextYLeft) = (a.x + leftBranch[0], a.y + leftBranch[1])
                    if self.gridmanager.in_bounds(nextXLeft, nextYLeft) and self.gridmanager(nextXLeft, nextYLeft).state == GridElementState.ROCK:
                        self.expand_crack(a.x, a.y, nextXLeft, nextYLeft, leftBranch)                
                        self.apicals.append(self.gridmanager(nextXLeft, nextYLeft))                    
                    rightBranch = rotate_right(a.direction_vector)
                    (nextXRight, nextYRight) = (a.x + rightBranch[0], a.y + rightBranch[1])
                    if self.gridmanager.in_bounds(nextXRight, nextYRight) and self.gridmanager(nextXRight, nextYRight).state == GridElementState.ROCK:
                        self.expand_crack(a.x, a.y, nextXRight, nextYRight, rightBranch) 
                        self.apicals.append(self.gridmanager(nextXRight, nextYRight))
                    self.apicals.remove(a)
                    self.unstables.append(a)
                else:
                    caveEventLogger.debug( "No.")

class Renderer(object):
    def __init__(self, surf, cellw, cellh, originx=0, originy=0):
        self.surface = surf
        self.cellw = cellw
        self.cellh = cellh
        self.originx = originx
        self.originy = originy
    def write_status_line(self, text, lineHeight):
        renderedText = infoFont.render(text, False, white)
        self.surface.fill(c_rock, pygame.Rect(0, 0, self.surface.get_width(), lineHeight))
        self.surface.blit(renderedText, (0,0))
    def transform(self, x, y):
        return (self.originx + x * self.cellw, self.originy + y * self.cellh)
    def fill_cell(self, x,y, color):
        (viewX, viewY) = self.transform(x,y)
        self.surface.fill(color, pygame.Rect(viewX, viewY, self.cellw, self.cellh))
    def blit_to_cell(self, x, y, surf):
        self.surface.blit(surf, self.transform(x,y))

# Game state setup
class PlayerState(object):
    def on_mushroom(self):
        self.health += 10
        messageQueue.append("That *was* a tasty mushroom! ...")
    def on_gem(self):
        self.nGems += 1
        chimeSound.play()
        messageQueue.append("Gem found! ...")
    def __init__(self, x, y, nGems, health, caveMgr):
        self.x = x
        self.y = y
        self.nGems = nGems
        self.health = health
        self.caveMgr = caveMgr
        self.alive = True
        self.handlers = { '*' : self.on_gem, 'm': self.on_mushroom }
    def try_move(self, gridmgr, dx, dy):
        if gridmgr.in_bounds(self.x + dx, self.y+dy) and gridmgr(self.x+ dx, self.y+ dy).state != GridElementState.ROCK:
            (self.x, self.y) = self.x + dx, self.y + dy
            recompute_visibility(self.x, self.y)
            self.health -= 1
            if self.health == 0:
                pygame.event.post(pygame.event.Event(DEATHEVENT, { 'cause': 'You ran out of health.'}))
        else:
             pygame.event.post(pygame.event.Event(FAILEDMOVEEVENT, {}))
    def try_get(self, gridmgr):
        playerEventLogger.debug("try get")
        self.caveMgr.iterate_over_unstables()
        global messageQueue
        result = gridmgr(self.x, self.y).try_get()
        if result: # None means there was nothing there.
            self.handlers[result]()
        else:
            messageQueue.append("Nothing here. ...")

def recompute_visibility(x, y, limit = 10):
    '''Recompute what is visible from (x,y) in all cardinal directions'''
    playerEventLogger.debug("Recomputing visibility at (%d, %d)" % (x,y))
    for v in direction_vectors:
        playerEventLogger.debug(" .. looking in direction " + str(v))
        (currX, currY) = (x,y)
        dist = 0
        playerEventLogger.debug(gridmgr(currX, currY))
        while gridmgr.in_bounds(currX, currY) and gridmgr(currX,currY).state != GridElementState.ROCK and dist < limit:
            playerEventLogger.debug("    .... at (%d,%d).. " % (currX, currY))
            gridmgr(currX,currY).visible = True
            (currX, currY) = (currX + v[0], currY + v[1])
            playerEventLogger.debug("    then move to %d, %d" % (currX, currY))
            dist += 1

# Gameplay parameters
caveGrowthInterval = 500
repeatMoveInterval = 250
hungerInterval = 3000
startingHealth = 150


## Main game management objects
def create_new_game(w, h):
    global gridmgr, cavegrowthmgr, renderer, messageQueue, player
    gridmgr = GridManager(w, h)
    gridmgr.run_initializer(lambda x,y: GridElementState(GridElementState.ROCK, x, y))
    cavegrowthmgr = CaveGrowthManager(gridmgr)
    cavegrowthmgr.starting_point(4)
    renderer = Renderer(screen, objectsurfaces['@'].get_width(), objectsurfaces['@'].get_height(), 0, statusLineHeight)
    messageQueue = []
    player = PlayerState(winWidth/2, winHeight/2, 0, startingHealth, cavegrowthmgr)

# Event timers
pygame.time.set_timer(CAVEGROWTHEVENT, caveGrowthInterval)
pygame.time.set_timer(REPEATEDMOVEEVENT, 0)
pygame.time.set_timer(HUNGEREVENT, hungerInterval)

def blit_centred_onto(surface, destSurface, vOffset):
    surface.blit(destSurface, ((surface.get_width() - destSurface.get_width())/2, vOffset))
    return destSurface.get_height()

def display_titlescreen():
    currVOffset = 100
    currVOffset += 10 + blit_centred_onto(screen, title, currVOffset)
    currVOffset += 10 + blit_centred_onto(screen, copyleftImage, currVOffset)
    currVOffset += 10 + blit_centred_onto(screen, creditText, currVOffset)
    currVOffset += 10 + blit_centred_onto(screen, pressToPlayText, currVOffset)
    for i in instructionsTextList:
        currVOffset += 10 + blit_centred_onto(screen, i, currVOffset)

def display_gameoverscreen():
    screen.fill(c_rock)
    causeOfDeathText = infoFont.render(messageQueue[0], False, white)
    finalScoreText = infoFont.render(messageQueue[1], False, white)
    currVOffset = 100
    currVOffset += 10 + blit_centred_onto(screen, gameOverText, currVOffset)
    currVOffset += 10 + blit_centred_onto(screen, causeOfDeathText, currVOffset)
    currVOffset += 20 + blit_centred_onto(screen, finalScoreText, currVOffset)
    for i in range(len(soundCreditsTextList)):
        spacing = 10 + (10 * ((i+1)%2))
        currVOffset += spacing + blit_centred_onto(screen, soundCreditsTextList[i], currVOffset)
    
def main():
    global gameState, messageQueue
    while (True):
        if gameState == STARTSCREEN:
            display_titlescreen()           
            events = pygame.event.get()
            for event in events: 
              if event.type == QUIT: 
                 sys.exit(0)
              elif event.type == KEYDOWN:
                 if event.key == K_q:
                     sys.exit(0)
                 create_new_game(winWidth, winHeight)
                 gameState = MAINSCREEN
            pygame.display.flip()
        elif gameState == ENDGAMESCREEN:
           display_gameoverscreen()
           events = pygame.event.get()
           for event in events: 
                if event.type == QUIT: 
                    sys.exit(0)
                elif event.type == KEYDOWN:
                     if event.key == K_q:
                         sys.exit(0)
                     create_new_game(winWidth, winHeight)
                     gameState = MAINSCREEN
           pygame.display.flip()
        else:
            if messageQueue:
                renderer.write_status_line(messageQueue[0], statusLineHeight)
            else:
                renderer.write_status_line('Gems Collected: %d      Health: %d' % (player.nGems, player.health), statusLineHeight)
            for x in range(gridmgr.w):
                for y in range(gridmgr.h):
                    curr = gridmgr(x,y)
                    if curr.state == GridElementState.STABLE:
                        renderer.fill_cell(x, y, c_stable)
                    elif curr.state == GridElementState.APICAL:
                        if curr.visible:
                            renderer.fill_cell(x, y, c_apical_visible)
                        else:
                            renderer.fill_cell(x, y, c_apical_invisible)
                    elif curr.state == GridElementState.UNSTABLE:
                        if curr.visible:
                            renderer.fill_cell(x,y, c_unstable_visible)
                        else:
                            renderer.fill_cell(x, y, c_unstable_invisible)
                    else:            
                        renderer.fill_cell(x,y, c_rock)
                    if curr.visible and curr.contents:
                        renderer.blit_to_cell(x, y, objectsurfaces[curr.contents[0]])
            renderer.blit_to_cell(player.x, player.y, objectsurfaces['@'])
            events = pygame.event.get()
            for event in events: 
              global last_key
              if event.type == QUIT: 
                 sys.exit(0)
              if event.type == DEATHEVENT:
                 messageQueue = [ event.cause, 'Gems Collected: ' + str(player.nGems) ]
                 #pygame.time.set_timer(CAVEGROWTHEVENT, 0)
                 gameState = ENDGAMESCREEN
              if event.type == KEYDOWN:
                 # Movement
                 if event.key == K_LEFT:
                     player.try_move(gridmgr, -1, 0,)
                 elif event.key == K_RIGHT:
                     player.try_move(gridmgr, 1, 0)
                 elif event.key == K_UP:
                     player.try_move(gridmgr, 0, -1)
                 elif event.key == K_DOWN:
                     player.try_move(gridmgr, 0, 1)
                 # Actions
                 elif event.key == K_PERIOD:    # try to pick up
                     player.try_get(gridmgr)
                 elif event.key == K_SPACE:    # move through status updates
                     messageQueue = messageQueue[1:]
                 elif event.key == K_q:
                     sys.exit(0)
                 last_key = event.key
                 pygame.time.set_timer(REPEATEDMOVEEVENT, repeatMoveInterval)
              if event.type == KEYUP:
                 pygame.time.set_timer(REPEATEDMOVEEVENT, 0)
              if event.type == REPEATEDMOVEEVENT:
                  if last_key == K_LEFT:
                      player.try_move(gridmgr, -1, 0)
                  elif last_key == K_RIGHT:
                      player.try_move(gridmgr, 1, 0)
                  elif last_key == K_UP:
                      player.try_move(gridmgr, 0, -1)
                  elif last_key == K_DOWN:
                      player.try_move(gridmgr, 0, 1)
              if event.type == FAILEDMOVEEVENT:
                 bonkSound.play()
              if event.type == CAVEGROWTHEVENT:
                 cavegrowthmgr.iterate_over_apicals()
                 recompute_visibility(player.x, player.y)
              if event.type == CAVEINEVENT:
                 caveinSound.play()
                 screen.scroll(5, 5)
                 if event.x == player.x and event.y == player.y:
                    pygame.event.post(pygame.event.Event(DEATHEVENT, { 'cause': 'You were caught in a cave-in.'}))
              if event.type == HUNGEREVENT:
                 player.health -= 1
            pygame.display.flip()

main()

